/************************************************************************/
/* ascii.c                                                              */
/*                                                                      */
/* Reading ASCII files.                                                 */
/*                                                                      */
/* Author: Bernhard Mayer,                                              */
/*        Fraunhofer Institute for Atmospheric Environmental Research,  */
/*        82467 Garmisch-Partenkirchen,                                 */
/*        Germany                                                       */
/*                                                                      */
/*----------------------------------------------------------------------*/
/* Copyright (C) 1995 Bernhard Mayer                                    */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*----------------------------------------------------------------------*/
/************************************************************************/


#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <ascii.h>


/***********************************/
/* Character to indicate a comment */
/***********************************/

#define ASCII_COMMENT_1          '%'
#define ASCII_COMMENT_2          '#'




/************************************/
/* Prototypes of internal functions */
/************************************/

static int ASCII_comment (char t);



/*****************************************************************************/
/* check an ASCII file; count rows, minimal number of columns,               */
/* maximal number of columns and the maximal length of a string;             */
/* empty rows and rows beginning with a commentar symbol (ASCII_COMMENT_1/2) */
/* are not counted.                                                          */ 
/*****************************************************************************/

int ASCII_checkfile (char *filename, 
		     int *rows,
		     int *min_columns,
		     int *max_columns,
		     int *max_length)
   
{
  FILE *f=NULL;
  char line[MAX_LENGTH_OF_LINE+1]="";
  char *token=NULL;
  char *string=NULL;
  int temp1=0, temp2=0;
  int min_col=INT_MAX, max_col=0, max_len=0, r=0;
  

  /* reset parameters */
  *min_columns=0;
  *max_columns=0;
  *max_length=0;
  *rows=0;

  string = line;

  
  if ( (f = fopen(filename, "r")) == NULL)  
    return ASCIIFILE_NOT_FOUND;
  

  /* count rows and columns */
  while ( fgets (string, MAX_LENGTH_OF_LINE, f) != NULL )  {

    if ( (token = strtok (string, " \t\n")) != NULL ) { /* if not an empty line */ 
      /* if not a comment     */
      if (!ASCII_comment(token[0]))  {

	r++;
	temp1++;

	/* check for maximal string length */
	temp2 = strlen(token);
	max_len = (temp2>max_len ? temp2 : max_len);

	while ( (token = strtok (NULL, " \t\n")) != NULL)  {

	  /* check for maximal string length */
	  temp2 = strlen(token);
	  max_len = (temp2>max_len ? temp2 : max_len);

	  if (ASCII_comment(token[0]))  /* if comment */
	    break;
	  temp1++;
	}

	min_col = (temp1<min_col ? temp1 : min_col);
	max_col = (temp1>max_col ? temp1 : max_col);
	temp1 = 0;
      }
    }

    string = line;
  }

  *min_columns = min_col;
  *max_columns = max_col;
  *max_length  = max_len;
  *rows        = r; 
  

  fclose (f);
  return 0;
}




/****************************************************************************/
/* allocate memory for a twodimensional string array                        */
/****************************************************************************/

int ASCII_calloc_string (char ****string,
			 int rows,
			 int columns,
			 int length)
{
  int i=0, j=0;

  if ((*string = (char ***) calloc (rows, sizeof (char **))) == NULL)
    return ASCII_NO_MEMORY;

  for (i=0; i<rows; i++)  {
    if (((*string)[i] = (char **) calloc (columns, sizeof (char *))) == NULL)
      return ASCII_NO_MEMORY;

    for (j=0; j<columns; j++)
      if (((*string)[i][j] = (char *) calloc (length+1, sizeof (char))) == NULL) 
	return ASCII_NO_MEMORY;
    
  }
  
  return 0;
}    




/****************************************************************************/
/* free memory, which was allocated with ASCII_calloc_string                */
/****************************************************************************/

int ASCII_free_string (char ***string, 
		       int rows,
		       int columns)

{
  int i=0, j=0;

  for (i=0; i<rows; i++)  {
    for (j=0; j<columns; j++)
      free (string[i][j]);

    free (string[i]);
  }

  free (string);

  return 0;
}    




/****************************************************************************/
/* allocate memory for a twodimensional double array                        */
/****************************************************************************/

int ASCII_calloc_double (double ***value,
			 int rows,
			 int columns)
{
  int i=0;
  
  if ( (*value = (double **) calloc ((size_t) rows, sizeof (double *))) == NULL )
    return ASCII_NO_MEMORY;

  for (i=0; i<rows; i++)  
    if ( ((*value)[i] = (double *) calloc ((size_t) columns, 
					   sizeof (double))) == NULL )
      return ASCII_NO_MEMORY;

  return 0;
}    



/****************************************************************************/
/* allocate memory for a twodimensional float array                         */
/****************************************************************************/

int ASCII_calloc_float  (float ***value,
			 int rows,
			 int columns)
{
  int i=0;

  if ( (*value = (float **) calloc (rows, sizeof (float *))) == NULL )  
    return ASCII_NO_MEMORY;

    
  for (i=0; i<rows; i++) 
    if ( ((*value)[i] = (float *) calloc (columns, sizeof (float))) == NULL )
      return ASCII_NO_MEMORY;

  return 0;
}    



/****************************************************************************/
/* free memory, which was allocated with ASCII_calloc_double                 */
/****************************************************************************/

int ASCII_free_double (double **value, 
		       int rows)
{
  int i=0;

  for (i=0; i<rows; i++) 
    free (value[i]);

  free (value);

  return 0;
}    


/****************************************************************************/
/* free memory, which was allocated with ASCII_calloc_float                 */
/****************************************************************************/

int ASCII_free_float (float **value, 
		       int rows)
{
  int i=0;

  for (i=0; i<rows; i++) 
    free (value[i]);

  free (value);

  return 0;
}    






/****************************************************************************/
/* read an ASCII file into a twodimensional array of strings;               */
/* before calling ASCII_readfile, you must check the file with              */
/* ASCII_checkfile and allocate memory with ASCII_calloc_string             */
/****************************************************************************/

int ASCII_readfile (char *filename, 
		    char ***array)
{
  FILE *f=NULL;
  char line[MAX_LENGTH_OF_LINE+1]="";
  char *string=NULL;
  char *t=NULL;
  int row=0, column=0;

  
  string = line;

  
  if ( (f = fopen(filename, "r")) == NULL)  
    return ASCIIFILE_NOT_FOUND;
  

  while ( fgets (string, MAX_LENGTH_OF_LINE, f) != NULL )  {

    column=0;

    if ( (t = strtok (string, " \t\n") ) != NULL)  {  /* if not an empty line */ 
      if (!ASCII_comment(t[0]))  { /* if not a comment */
	strcpy (array[row][column++], t);
	
	while ( (t = strtok (NULL, " \t\n") ) != NULL)  {
	  if (ASCII_comment(t[0]))     /* if comment */
	    break;
	  strcpy (array[row][column++], t);
	}

	row++;
      }
    }

    string = line;
  }


  fclose (f);
  return 0;
}



/****************************************************************************/
/* convert string array to double array                                     */
/****************************************************************************/

int ASCII_string2double (double **value, 
			 char *** string,
			 int rows,
			 int columns)
{
  int i=0, j=0;
  char *dummy=NULL;

  for (i=0; i<rows; i++) 
    for (j=0; j<columns; j++)  {

      if (string[i][j][0] == 0)  
	value[i][j] = NAN;
      else  
	value[i][j] = strtod (string[i][j], &dummy);

    }

  return 0;
}    




/****************************************************************************/
/* convert string array to float array                                      */
/****************************************************************************/

int ASCII_string2float (float **value, 
			char *** string,
			int rows,
			int columns)
{
  int i=0, j=0;
  char *dummy=NULL;

  for (i=0; i<rows; i++)  { 
    for (j=0; j<columns; j++)  {
      if (string[i][j][0] == 0)   {
	value[i][j] = (float) NAN;
      }
      else  {
	value[i][j] = (float) strtod (string[i][j], &dummy);
      }
    }
  }
  return 0;
}    







/****************************************************************************/
/*                                                                          */
/* read an ASCII file; values are stored in the twodimensional array        */
/* value [row] [column]; memory allocation for value is done automatically. */
/* rows is the number of (not empty) rows of the file, max_columns is the   */
/* maximal number of columns of the file and min_columns is the minimal     */
/* number of columns of all (not empty) lines; the dimension of the array   */
/* value is rows * max_columns; strings that cannot be interpreted as       */
/* floating point number are converted to 0; rows with less than            */
/* max_columns columns are filled up with NAN; the memory allocated by      */
/* input_file can be freed with ASCII_free_double (&value, rows).           */
/*                                                                          */
/****************************************************************************/

int ASCII_file2double (char *filename,
		       int *rows,
		       int *max_columns,
		       int *min_columns,
		       double ***value)
{
  char ***string = NULL;
  int status=0;
  int max_length=0;

  /* count rows and columns of ASCII file <filename> */
  if ( (status = ASCII_checkfile (filename, 
				  rows, 
				  min_columns, 
				  max_columns, 
				  &max_length)) != 0)
    return status;
  


  /* allocate memory for string array */
  if ( (status = ASCII_calloc_string (&string, 
				      *rows, 
				      *max_columns, 
				      max_length)) != 0 )
    return status;
  
  /* read ASCII file to string array */
  if ( (status = ASCII_readfile (filename, string)) != 0)
    return status;

  /* allocate memory for double array */
  if ( (status = ASCII_calloc_double (value, *rows, *max_columns)) != 0 )
    return status;

  /* convert string array to double array */
  if ( (status = ASCII_string2double (*value, string, *rows, *max_columns)) != 0)
    return status; 



  /* free memory of string array */
  ASCII_free_string (string, *rows, *max_columns);

  return 0;  /* everything ok */
} 



/****************************************************************************/
/*                                                                          */
/* read an ASCII file; values are stored in the twodimensional array        */
/* value [row] [column]; memory allocation for value is done automatically. */
/* rows is the number of (not empty) rows of the file, max_columns is the   */
/* maximal number of columns of the file and min_columns is the minimal     */
/* number of columns of all (not empty) lines; the dimension of the array   */
/* value is rows * max_columns; strings that cannot be interpreted as       */
/* floating point number are converted to 0; rows with less than            */
/* max_columns columns are filled up with NAN; the memory allocated by      */
/* input_file can be freed with ASCII_free_float (&value, rows).            */
/*                                                                          */
/****************************************************************************/

int ASCII_file2float (char *filename,
		      int *rows,
		      int *max_columns,
		      int *min_columns,
		      float ***value)
{
  char ***string = NULL;
  int status=0;
  int max_length=0;

  /* count rows and columns of ASCII file <filename> */
  if ( (status = ASCII_checkfile (filename, 
				  rows, 
				  min_columns, 
				  max_columns, 
				  &max_length)) != 0)
    return status;
  

  /* allocate memory for string array */
  if ( (status = ASCII_calloc_string (&string, 
				      *rows, 
				      *max_columns, 
				      max_length)) != 0 )
    return status;
  

  
  /* read ASCII file to string array */
  if ( (status = ASCII_readfile (filename, string)) != 0)
    return status;


  /* allocate memory for float array */
  if ( (status = ASCII_calloc_float (value, *rows, *max_columns)) != 0 )
    return status;

  /* convert string array to float array */
  if ( (status = ASCII_string2float (*value, string, *rows, *max_columns)) != 0)
    return status; 

  /* free memory of string array */
  ASCII_free_string (string, *rows, *max_columns);

  
  return 0;  /* everything ok */
} 



/****************************************************/
/*                                                  */
/* get a column of a double array; allocate memory; */
/* return pointer to column                         */
/*                                                  */
/****************************************************/

double *ASCII_column (double **value, int rows, int column)  
{
  int i=0;
  double *col = (double *) calloc (rows, sizeof(double));

  for (i=0; i<rows; i++)
    col[i] = value[i][column];

  return col;
}
  
/****************************************************/
/*                                                  */
/* get a column of a float array; allocate memory; */
/* return pointer to column                         */
/*                                                  */
/****************************************************/

float *ASCII_column_float (float **value, int rows, int column)  
{
  int i=0;
  float *col = (float *) calloc (rows, sizeof(float));

  for (i=0; i<rows; i++)
    col[i] = value[i][column];

  return col;
}
  




/***********************************************************************/
/* Read an ASCII file with (at least) 1 column.                        */
/* Only the first column is returned in array first;                   */ 
/* n is the number of values returned.                                 */
/* memory allocation for first is done automatically and               */
/* can be freed with an ordinary free().                               */ 
/***********************************************************************/

int read_1c_file (char *filename, 
		  double **first, int *n)
{
  int max_columns=0, min_columns=0;
  double **data=NULL;
  int status=0;

  /* read file */
  if ( (status = ASCII_file2double (filename, n, 
				    &max_columns, &min_columns, &data)) != 0)
    return status;


  /* check, if at least two columns */
  if (min_columns < 1)   {
    ASCII_free_double (data, *n);
    return LESS_THAN_TWO_COLUMNS;
  }
  
  *first  = ASCII_column (data, *n, 0);
  
  ASCII_free_double (data, *n);

  return 0;
}




/***********************************************************************/
/* Read an ASCII file with (at least) 2 columns.                       */
/* Only the first two columns are returned in arrays first and second; */ 
/* n is the number of values returned.                                 */
/* memory allocation for first and second is done automatically and    */
/* can be freed with an ordinary free().                               */ 
/***********************************************************************/

int read_2c_file (char *filename, 
		  double **first, double **second, int *n)
{
  int max_columns=0, min_columns=0;
  double **data=NULL;
  int status=0;

  /* read file */
  if ( (status = ASCII_file2double (filename, n, 
				    &max_columns, &min_columns, &data)) != 0)
    return status;


  /* check, if at least two columns */
  if (min_columns < 2)   {
    ASCII_free_double (data, *n);
    return LESS_THAN_TWO_COLUMNS;
  }
  
  *first  = ASCII_column (data, *n, 0);
  *second = ASCII_column (data, *n, 1);
  
  ASCII_free_double (data, *n);

  return 0;
}



/***********************************************************************/
/* Read an ASCII file with (at least) 3 columns.                       */
/* Only the first three columns are returned in arrays first, second   */ 
/* and third.                                                          */
/* n is the number of values returned.                                 */
/* memory allocation for first and second is done automatically and    */
/* can be freed with an ordinary free().                               */ 
/***********************************************************************/

int read_3c_file (char *filename, 
		  double **first, double **second, double **third, int *n)
{
  int max_columns=0, min_columns=0;
  double **data=NULL;
  int status=0;

  /* read file */
  if ( (status = ASCII_file2double (filename, n, 
				    &max_columns, &min_columns, &data)) != 0)
    return status;


  /* check, if at least three columns */
  if (min_columns < 3)   {
    ASCII_free_double (data, *n);
    return LESS_THAN_THREE_COLUMNS;
  }
  
  *first  = ASCII_column (data, *n, 0);
  *second = ASCII_column (data, *n, 1);
  *third  = ASCII_column (data, *n, 2);
  
  ASCII_free_double (data, *n);

  return 0;
}




/***********************************************************************/
/* Read an ASCII file with (at least) 5 columns.                       */
/* Only the first five columns are returned in arrays first, second,   */ 
/* third, fourth and fifth.                                            */
/* n is the number of values returned.                                 */
/* memory allocation for first and second is done automatically and    */
/* can be freed with an ordinary free().                               */ 
/***********************************************************************/

int read_5c_file (char *filename, 
		  double **first, double **second, double **third, 
		  double **fourth, double **fifth, int *n)
{
  int max_columns=0, min_columns=0;
  double **data=NULL;
  int status=0;

  /* read file */
  if ( (status = ASCII_file2double (filename, n, 
				    &max_columns, &min_columns, &data)) != 0)
    return status;


  /* check, if at least three columns */
  if (min_columns < 5)   {
    ASCII_free_double (data, *n);
    return LESS_THAN_THREE_COLUMNS;
  }
  
  *first  = ASCII_column (data, *n, 0);
  *second = ASCII_column (data, *n, 1);
  *third  = ASCII_column (data, *n, 2);
  *fourth = ASCII_column (data, *n, 3);
  *fifth  = ASCII_column (data, *n, 4);
  
  ASCII_free_double (data, *n);

  return 0;
}




/*****************************************************************/
/* Create substring starting at position "start" with length     */
/* "length". Result is written to buffer (which MUST be          */
/* allocated before). "buffer" is returned.                      */  
/*****************************************************************/

char *substr(char *buffer, char *string, int start, int length)
{
  strncpy (buffer, string+start, length);                    
  buffer[length]=0;  /* end of string */

  return buffer;
}





/*****************************************************************/
/* Parse string to an array of single words. Memory for an array */
/* of string pointers is allocated automatically. array[i]       */
/* points to the address of word #i in string!                   */
/* Word separator is specified in separator.                     */
/* Number of words is returned in number.                        */
/* Characters following the comment character are ignored.       */
/*****************************************************************/

int ASCII_parse (char *string, char *separator, char ***array, int *number)
{

  char *start=NULL;
  char *t=NULL;
  char *save=NULL;
  char **temp=NULL;

  /* save start address of string */
  start = string;

  /* save string */
  save = (char *) calloc (strlen(string)+1, sizeof (char));
  strcpy (save, string);


  /* reset number */
  *number=0;
  
  /* count words */
  if ( (t = strtok (string, separator) ) != NULL)  {  /* if not an empty line */ 
    if (!ASCII_comment(t[0]))  {  /* if not a comment     */
      (*number)++;
      while ( (t = strtok (NULL, separator) ) != NULL)  {
	if (ASCII_comment(t[0]))     /* if comment */
	  break;

	(*number)++;
      }
    }
  }

  if (*number==0)   {
    free(save);
    return 0;     /* no words found. but o.k. */ 
  }

  /* restore string */
  string = start;

  /* allocate memory for an array of *number character pointers */
  temp = (char **) calloc (*number, sizeof(char *));

  /* restore string */
  strcpy (string, save);

  /* reset *number */
  *number=0;  

  /* now set array pointers */
  if ( (t = strtok (string, separator) ) != NULL)  {
    if (!ASCII_comment(t[0]))  {             
      temp[(*number)++] = t;
      
      while ( (t = strtok (NULL, separator) ) != NULL)  {
	if (ASCII_comment(t[0]))    
	  break;

	temp[(*number)++] = t;
      }
    }
  }


  /* copy temp to array */
  *array = temp;

  free(save);
  return 0;   /* if o.k. */
}



/*****************************************************************/
/* For compatibility reasons: ASCII_parsestrings() is just       */
/* ASCII_parse() with separator set to " \t\n"                   */
/*****************************************************************/

int ASCII_parsestring (char *string, char ***array, int *number)
{
  int status=0;

  status = ASCII_parse (string, " \t\n", array, number);
  return status;
}



/*****************************************************************/
/* check if an ASCII character is a comment char                 */
/*****************************************************************/

static int ASCII_comment (char t) {

  if (t == ASCII_COMMENT_1 || t == ASCII_COMMENT_2)
    return 1;
  else 
    return 0;
}
