      SUBROUTINE crs_o3( crs, error, first, lambda, n_lambda, ntemp, 
     $     path, temper, type)  
*
*BOD
*DESCRIPTION
* Calculates ozone cross sections as a function of
* wavelength and temperature.
*USAGE
**FORTRAN
*     CALL crs_o3( crs, error, lambda, n_lambda, ntemp, path,
*    $             temper, type)  
**PERL
*INPUT
**first
*     logical
*      If true, temperature coefficients for ozone will be 
*      calculated at wavelengths lambda. If first is true, it will
*      be returned as false, just in case you shold forget.
*      If false, temperature coefficients from earlier call will be
*      used. 
*      first should be true the first time crs_o3 is called. It should 
*      also be true if lambda is changed. It does not need to be true
*      if temper is changed.
**lambda
*     real array(*)
*     Wavelengths (nm) at which the ozone cross section is to be calculated.
**n_lambda
*     integer
*     Number of wavelengths lambda.
**ntemp
*     integer
*     Number of different temperatures (levels in the atmosphere) 
**path
*     character string
*     The location of the Radtran data files, e.g. 
*     /usr/local/lib/Radtran/data
**temper
*     real array(*)
*     The temperatures (levels in the atmosphere) at which
*     the ozone cross section is to be calculated.
**type
*     integer
*     The ozone cross section data to be used:
***1     Molina and Molina (1986)
***2     Daumont (not implemented yet)
*OUTPUT
**crs
*     real array(temper,lambda)
*     The ozone cross section as a function of temper and lambda.
**error
*     integer
*     Error flag. Error checking performed only if first is true.
***0     Everything ok.
***1     Error occurred during reading of data file.
***2     n_lambda == 0 and ntemp != 0
***3     n_lambda != 0 and ntemp == 0
*ERROR_MESSAGES
*ROUTINES_CALLED
*FILES
*     crs_o3_Molina.dat        Data file
*     consts.inc               Include file
*     param.inc                Include file
*REFERENCES
**      Molina and Molina, 'Absolute Absorption Cross Sections of
*        Ozone in the 185- to 350-nm wavelength Range', JGR, vol 91,
*        pp. 14,501-14,508, 1986.
*
*EOD
*
*
*     References:
*     
*     WMO (1986), 'Atmospheric Ozone 1985', WMO Report No. 16.
*
*     
      INCLUDE 'param.inc'
*
      CHARACTER*(*) path
      INTEGER  error,  n_lambda, ntemp, type
      LOGICAL first
*
      REAL crs(0:maxcly, maxpro, *), lambda(*), temper(0:*)
*
* Internal variables
*     
      CHARACTER*1  char
      CHARACTER*18 filenm    ! No. of characters must equal
                             ! length of filename.
      CHARACTER*100 fullname, line
      INTEGER n_temp, n_temp_raw(maxwvn), n_crs
      REAL crs_temp(NCRSTEMP), crs_raw(NCRSTEMP,maxwvnraw), fact,
     $     lambda_raw(maxwvnraw), crs_new(NCRSTEMP),
     $     temp_new(NCRSTEMP,maxwvn), coeff(NCRSTEMP, maxwvn)
*
      INTEGER closest_above, closest_below
*
      SAVE  n_temp, n_crs, crs_temp, crs_raw, lambda_raw, coeff,
     $     n_temp_raw, temp_new
*
      INCLUDE 'consts.inc'
*
      error = 0
      IF ( first ) THEN
*
         CALL zeroit( crs, (maxcly+1)*maxwvn*maxpro)
*
         IF ( ( n_lambda .LE. 0 ) .AND. ( ntemp .NE. 0 ) ) THEN
            error = 2
            RETURN
         ELSE IF ( ( n_lambda .LE. 0 ) .AND. ( ntemp .NE. 0 ) ) THEN
            error = 3
            RETURN
         ENDIF
*
* Read in all data on first entry only
*
         IF ( type .EQ. 1 ) THEN
*            filenm   = 'crs/crs_o3_Molina.dat'
            filenm   = 'crs/crs_o3_all.dat'
            n_temp   = 3
            n_crs    = 1
            fact     = 1.E-20
         ENDIF
         fullname = path//filenm
      write(*,*) "inside crs_o3", fullname
         OPEN(UNIT=iunit,FILE=fullname,STATUS='OLD',
     $        FORM='FORMATTED',ERR=99, IOSTAT=ios)
         GOTO 991
 99      CONTINUE
         WRITE(*,'(A,A)') 'Error during read of file: ', fullname
         WRITE(*,'(A, i4)') 'IOSTAT: ', ios
         error = 1
         RETURN
 991     CONTINUE
*     
      write(*,*) "reading ", fullname
         DO WHILE ( .TRUE. ) 
            READ(iunit,'(A)',END=98 ) line
      write(*,*) "reading ", line
*
* Read all the no-comment lines
*
            IF ( line(:1) .NE. comcha ) THEN
               IF ( type .EQ. 1 ) THEN
                  IF ( line(:1) .EQ. tempcha ) THEN
                     READ(line,*) char, (crs_temp(it),it=1, n_temp)
                  ELSE
                     READ(line,*) lambda_raw(n_crs),   
     $                    (crs_raw(it, n_crs),it=1, n_temp)
                     DO it=1, n_temp
                        crs_raw(it, n_crs) = crs_raw(it, n_crs) * fact
                     ENDDO
                        write(*,*)"raw",
     $                    (crs_raw(it, n_crs),it=1, n_temp)
                     n_crs = n_crs + 1
                  ENDIF
               ENDIF
            ENDIF
         ENDDO
 98      CONTINUE
      write(*,*) "done reading ", n_crs, line
         n_crs = n_crs - 1
         CLOSE(iunit)
*
* Return raw cross section if n_lambda and ntemp equal zero
*
         IF ( ( n_lambda .EQ. 0 ) .AND. ( ntemp .EQ. 0 ) ) THEN
            DO it = 0, n_temp
               temper(it) = crs_temp(it)
               DO iv = 1, n_crs
                  lambda(iv)  = lambda_raw(iv)
                  crs(it, io3, iv) = crs_raw(it,iv)
               ENDDO
            ENDDO
            n_lambda = n_crs
            ntemp    = n_temp
            RETURN
         ENDIF
*
* Calculate the temperature coefficients for all lambda.
*
         DO iv = 1, n_lambda
*
*  Find the closest wavelengths in original data.
*
*
*FIXME At the moment only cross section at lambda_raw may be calculated,
*      Include any lambda in the future.
*
            ia = closest_above(lambda(iv), lambda_raw, n_crs)
            ib = closest_below(lambda(iv), lambda_raw, n_crs)
*            write(*,*) "doing closest", iv, lambda(iv), ia,ib,
*     $           crs_raw(1,ia), crs_raw(1,ib)
            IF ( ia .EQ. ib ) THEN
               ir = ia
*
*  Figure out at how many temperatures the cross section 
*  is reported.
*
*     Molina and Molina report some values only at two temperatures.
               IF ( type .EQ. 1 .AND. crs_raw(2,ir) .EQ. 0.0 ) THEN
                  n_temp_raw(iv) = 2
                  crs_new(1) = crs_raw(1,ir)
                  crs_new(2) = crs_raw(3,ir)
                  temp_new(1,iv)= crs_temp(1)
                  temp_new(2,iv)= crs_temp(3)
               ELSE IF ( type .EQ. 1 ) THEN
                  n_temp_raw(iv) = 3
                  crs_new(1) = crs_raw(1,ir)
                  crs_new(2) = crs_raw(2,ir)
                  crs_new(3) = crs_raw(3,ir)
                  temp_new(1,iv)= crs_temp(1)
                  temp_new(2,iv)= crs_temp(2)
                  temp_new(3,iv)= crs_temp(3)
cakyp
*                  WRITE(*,*)"CRS",iv,lambda(iv),crs_new(1),
*     $                 crs_new(2),crs_new(3),
*     $                 temp_new(1,iv),temp_new(2,iv),temp_new(3,iv)
cakyp
               ENDIF
            ELSE IF ( (ib .EQ. 0) .AND. (ia .GT. n_crs) ) THEN
                  n_temp_raw(iv) = 3
                  crs_new(1) = 0.0
                  crs_new(2) = 0.0
                  crs_new(3) = 0.0
                  temp_new(1,iv)= crs_temp(1)
                  temp_new(2,iv)= crs_temp(2)
                  temp_new(3,iv)= crs_temp(3)               
            ELSE IF ( (crs_raw(1,ia) .EQ. 0.0) .AND. 
     $                 (crs_raw(1,ib) .EQ. 0.0) ) THEN
                  n_temp_raw(iv) = 0
                  crs_new(3) = 0.0
                  temp_new(1,iv)= 0.0
            ELSE
               WRITE(*,*)"crs_o3, can not handle this case yet"
            ENDIF
cakyp            write(*,*) "doing newton", iv, lambda(iv), maxwvn, maxwvnraw
*
*  Use these cross sections to calculate temperature coefficients.
*
            CALL newton_co(n_temp_raw(iv), temp_new(1,iv), crs_new, 
     $           coeff(1,iv))
cakyp
*                  WRITE(*,*)"coeff",iv,lambda(iv),coeff(1,iv)
cakyp
         ENDDO
*
         first = .FALSE.
      ENDIF
*
********** Done with first ************************************
*
* Below follows what is done on .NOT. first calls
*
* Calculate ozone cross sections 
*
      DO iv = 1, n_lambda
         DO it = 0, ntemp
            crs(it, io3, iv) = eval(n_temp_raw(iv), temp_new(1,iv),
     $           coeff(1,iv), temper(it))
         ENDDO
         it = ntemp
            write(*,*) "loop crs_o3", iv, it, n_lambda, ntemp,
     $      lambda(iv), crs(it, io3, iv), 
     $        coeff(1,iv), coeff(2,iv), coeff(3,iv),
     $        n_temp_raw(iv), temp_new(1,iv), temp_new(2,iv), 
     $        temp_new(3,iv),
     $        temper(it)

      ENDDO
*
      RETURN
      END
      
