/************************************************************************/
/* stamntab.c                                                           */
/*                                                                      */
/* Calculate Stamnes table.                                             */
/*                                                                      */
/* Author: Bernhard Mayer,                                              */
/*        Fraunhofer Institute for Atmospheric Environmental Research,  */
/*        82467 Garmisch-Partenkirchen,                                 */
/*        Germany                                                       */
/*                                                                      */
/*----------------------------------------------------------------------*/
/* Copyright (C) 1997 Bernhard Mayer                                    */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*----------------------------------------------------------------------*/
/************************************************************************/


#define PROGRAM "STAMNTAB"
#define VERSION "0.9e"
#define DATE    "May 9, 1997"


#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
/* #include <process.h> */
#include <float.h>
#include <math.h>
#include <string.h>
#include <numeric.h>
#include <ascii.h>
/* #include <dbase.h> */
/* #include <sbase.h> */
#include <getopt.h>
/* #include <stamnes.h> */

#define DEFAULT_MODEL       TUVSPEC
#define DEFAULT_RESOLUTION  LORES3


#define DEFAULT_LOWER      305.0
#define DEFAULT_UPPER      340.0


/* prototypes of internal functions */
static void print_usage (char *filename);
static int get_options (int argc, char **argv,
			char *programname,
			char *stafilename,
			char *atmfilename,
			MODEL *model,
			RESOLUTION *res,
			STYLE *style,
			char *slitfunction,
			double *lambda_lower,
			double *lambda_upper,
			char *conv,
			char *diffuse);

static void print_status(char *programname,
			 char *stafilename,
			 char *atmfilename,
			 MODEL model,
			 RESOLUTION resolution,
			 STYLE style,
			 char *slitfunction,
			 double lambda_lower,
			 double lambda_upper,
			 char conv,
			 char diffuse);



/*************************************************************/
/* print usage information                                   */
/*************************************************************/

static void print_usage (char *filename)
{
  fprintf (stderr, "%s %s - Creating Stamnes table\n\n", PROGRAM, VERSION);
  fprintf (stderr, "written by  Bernhard Mayer,\n");
  fprintf (stderr, "            Fraunhofer-Institute for Atmospheric Research (IFU)\n");
  fprintf (stderr, "            82467 Garmisch-Partenkirchen, Germany\n");
  fprintf (stderr, "Version %s finished %s\n\n", VERSION, DATE);
  fprintf (stderr, "Be aware that this program is a beta version! Please report any\n");
  fprintf (stderr, "kind of error to the author, including the error message and the\n");
  fprintf (stderr, "database being processed when the error occured. Thanks!\n\n");
  fprintf (stderr, "USAGE: %s [options] <Stamnes table> <atmosphere>\n\n", filename);
  fprintf (stderr, "%s will calculate a so called Stamnes table to be used for \n", PROGRAM);
  fprintf (stderr, "calculation of ozone column from measured global UV spectra.\n\n");
  fprintf (stderr, "Necessary command line options:\n");
  fprintf (stderr, "  -l <lower>     lower wavelength [nm]\n"); 
  fprintf (stderr, "  -u <upper>     upper wavelength [nm]\n\n"); 
  fprintf (stderr, "Optional:\n");
  fprintf (stderr, "  -D  disort version (slow)\n");
  fprintf (stderr, "  -P  pseudo spherical version (slow)\n");
  fprintf (stderr, "  -S  spherical version (very slow)  \n");
  fprintf (stderr, "  -T  two stream version (default)   \n");
  fprintf (stderr, "  -B  model type Pseudo-Pseudospherical, alpha version!\n");
  fprintf (stderr, "  -A  STAR model of the University of Munich!\n");
  fprintf (stderr, "  -M  Sasha Madronich Model, alpha version!\n\n");
  fprintf (stderr, "  -E  use enhanced version with Angstrom alpha/beta and\n");
  fprintf (stderr, "      single scattering albedo.\n");
  fprintf (stderr, "  -O  old-fashioned, which means convolution before modelling ...\n\n");
  fprintf (stderr, "  -2  IFU system #2                  \n");
  fprintf (stderr, "  -3  IFU system #3 (default)        \n");
  fprintf (stderr, "  -q  <slitfunction file>  user-defined slitfunction\n\n");
  fprintf (stderr, "  -c  convolute model data with 5nm FWHM triangle\n");
  fprintf (stderr, "  -d  no direct sun\n");
}









/*************************************************************/
/* parse command line options                                */
/*************************************************************/

static int get_options (int argc, char **argv,
			char *programname,
			char *stafilename,
			char *atmfilename,
			MODEL *model,
			RESOLUTION *res,
			STYLE *style,
			char *slitfunction,
			double *lambda_lower,
			double *lambda_upper,
			char *conv,
			char *diffuse)
{
  char c=0;
  char enhanced=0;
  char *dummy=NULL;

  /* set default values */
  *lambda_lower = DEFAULT_LOWER;
  *lambda_upper = DEFAULT_UPPER;
  *model        = DEFAULT_MODEL;
  *res          = DEFAULT_RESOLUTION;
  *style        = BETTER;
  *conv         = 0;
  *diffuse      = 0;


  /* save name of program */
  strncpy (programname, argv[0], FILENAME_MAX);

  /* get command line arguments */
  optmode = GETOPT_ANY;

  while ((c=getopt (argc, argv, "DPSTBMAEO23q:l:u:cd")) != EOF)  {
    switch(c)  {
    case 'D':  /* disort */  
      *model = UVSPEC;
      break;   
    case 'P':  /* pseudo spherical */
      *model = PSUVSPEC;
      break;
    case 'S':  /* spherical */
      *model = SUVSPEC;
      break;
    case 'T':  /* twostr    */
      *model = TUVSPEC;
      break;
    case 'B':  /* pseudo pseudo spherical */
      *model = PSPSUVSPEC;
      break;
    case 'M':  /* Sasha Madronich model   */
      *model = SASHA_TWOSTR;
      break;
    case 'E':  /* enhanced version */
      enhanced = 1;
      break;
    case 'A':  /* STAR model   */
      *model = STAR;
      break;
    case 'O':  /* old-fashioned           */
      *style = OLD;
      break;
    case '2':  /* system #2 */
      *res = LORES2;
      break;
    case '3':  /* system #3 */
      *res = LORES3;
      break;
    case 'q':  /* user-defined slitfunction  */
      *res = SLIT;
      strcpy (slitfunction, optarg);
      break;
    case 'l':  /* lower wavelength */
      *lambda_lower = strtod (optarg, &dummy);
      break;
    case 'u':  /* upper wavelength */
      *lambda_upper = strtod (optarg, &dummy);
      break;
    case 'c':  /* convolute with 5nm triangle */
      *conv = 1;
      break;
    case 'd':  /* diffuse only */
      *diffuse = 1;
      break;
    case '?':
      print_usage (programname);
      return (-1);
      break;
    default:
      print_usage (programname);
      return (-1);
    }
  }
      

  /* check number of remaining command line arguments */
  if (argc - optind != 2)  {
    print_usage (programname);
    return (-1);
  }
  
  if (enhanced) {  /* check for enhanced model */
    switch (*model) {
    case TUVSPEC:
      *model = TUVSPEC_ENH;
      break;
    case UVSPEC:
      *model = UVSPEC_ENH;
      break;
    case PSUVSPEC:
      *model = PSUVSPEC_ENH;
      break;
    case SUVSPEC:
      *model = SUVSPEC_ENH;
      break;
    case PSPSUVSPEC:
      *model = PSPSUVSPEC_ENH;
      break;
    default:
      break;
    }
  }


  /* save command line arguments */
  strncpy (stafilename, argv[optind+0], FILENAME_MAX);
  strncpy (atmfilename, argv[optind+1], FILENAME_MAX);

  return 0;  /* if o.k. */
}
	  





/************************************************************/
/* Print status data.                                       */
/************************************************************/

static void print_status(char *programname,
			 char *stafilename,
			 char *atmfilename,
			 MODEL model,
			 RESOLUTION resolution,
			 STYLE style,
			 char *slitfunction,
			 double lambda_lower,
			 double lambda_upper,
			 char conv,
			 char diffuse)
{


  fprintf (stderr, " ... calculating Stamnes table for wavelengths %g and %gnm\n",
	   lambda_lower, lambda_upper);
  if(conv)
    fprintf (stderr, " ... convoluting with 5nm FWHM triangle\n"); 

  if(diffuse)
    fprintf (stderr, " ... no direct sun\n"); 


  switch(style) {
  case OLD:
    fprintf (stderr, " ... old-fashioned modelling, that is using convoluted extraterrestrial irradiance.\n"); 
    break;
  case BETTER:
    fprintf (stderr, " ... better modelling, that is HIRES calculation, convolution afterwards.\n"); 
    break;
  default:
    break;
  }

  switch(resolution) {
  case HIRES: 
    fprintf (stderr, " ... high resolution\n");
    break;
  case LORES2: 
    fprintf (stderr, " ... IFU mono #2\n");
    break;
  case LORES3: 
    fprintf (stderr, " ... IFU mono #3\n");
    break;
  case SLIT: 
    fprintf (stderr, " ... user-defined slitfunction %s\n", slitfunction);
    break;
  default:
    fprintf (stderr, " ... Hmm, unknown resolution\n");
    break;
  }
}




int main(int argc, char **argv)
{
  int status=0;

  char programname [FILENAME_MAX+1] = "";
  char stafilename [FILENAME_MAX+1] = "";
  char atmfilename [FILENAME_MAX+1] = "";

  MODEL      model = DEFAULT_MODEL;
  RESOLUTION resolution = DEFAULT_RESOLUTION;
  STYLE      style      = BETTER;

  char slitfunction[FILENAME_MAX+1] = "";

  double lambda_lower=0, lambda_upper=0;

  char conv=0, diffuse=0;

  /* allocate memory for atmospheric data */
  atmospheric_data *atm = (atmospheric_data *) calloc (1, sizeof(atmospheric_data));

  /* parse command line options */
  status = get_options (argc, argv,
			programname, stafilename,
			atmfilename, 
			&model,
			&resolution,
			&style,
			slitfunction,
			&lambda_lower,
			&lambda_upper,
			&conv,
			&diffuse);

  if (status!=0) 
    return status;


  /* print information about input data */
  print_status (programname, stafilename,
	        atmfilename, 
		model,
		resolution,
		style,
		slitfunction,
		lambda_lower, lambda_upper,
		conv, diffuse);
  
  fprintf (stderr, " ... reading atmospheric data file %s\n", atmfilename);
  
  /* read atmospheric data file */
  status = read_atmospheric_file (atmfilename, atm);
  
  if (status!=0)  {
    fprintf (stderr, "read_atmospheric_file(%s, ...) returned status %d\n", 
	     atmfilename, status);  
    return status;
  }


  /* print atmospheric data to stderr    */
  fprintf_atmospheric (stderr, atm);  


  /* create Stamnes table */
  status = create_stamnes_table (stafilename,
				 atm,
				 model, resolution, style,
				 slitfunction,
				 lambda_lower, lambda_upper,
				 conv, diffuse);
    
  if (status!=0) {
    fprintf (stderr, " ... error %d calculating Stamnes table\n", status);
    return status;
  }

  return 0;
}
