/*
 Copyright (C) 1997 Arve Kylling

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 1, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 To obtain a copy of the GNU General Public License write to the
 Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ascii.h>
#include "../src_c/getopt.h"
#include "../src_f/f77-uscore.h"
#include "uvspec.h"


int read_atmosphere(char* atmosphere_filename, atm_struct *atm) 
{
  int lev;
  int rows=0, min_columns=0, max_columns=0;
  int status = 0;
  float **atm_data=NULL;
  float *dens=NULL;
  
  /*  printf("%s\n",input.atmosphere_filename); */

  status = ASCII_file2float (atmosphere_filename, 
			      &rows, &max_columns, &min_columns, 
			      &atm_data);
  if (status!=0) {
    fprintf (stderr, " ... error %d opening file %s\n", status, atmosphere_filename);
    return status;
  }
  if (max_columns!=min_columns) {
    fprintf (stderr, " !! ATTENTION !! Inconsistent number of columns\n");
    fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
  }  
  if (min_columns<2) {
    fprintf (stderr, " ... ending, too few columns\n");
    return 0;
  }
  
  atm->nlev = rows;
  atm->nlyr = rows-1;
  
  atm->press     = ASCII_column_float(atm_data, atm->nlev, ATMFILE_PRESS );
  atm->temper    = ASCII_column_float(atm_data, atm->nlev, ATMFILE_TEMPER );
  atm->zd        = ASCII_column_float(atm_data, atm->nlev, ATMFILE_ZD );
  if ( (status = ASCII_calloc_float (&atm->dens, 2, atm->nlyr)) != 0 )
    return status;
  dens                = ASCII_column_float(atm_data, atm->nlev, ATMFILE_AIR );
  for (lev=0;lev<atm->nlev;lev++) { 
    atm->dens[AIR][lev] = dens[lev];
  }
  dens                = ASCII_column_float(atm_data, atm->nlev, ATMFILE_O3 );
  for (lev=0;lev<atm->nlev;lev++) { 
    atm->dens[O3][lev] = dens[lev];
  }
  free(atm_data);
  free(dens);
  return 0;
}

int set_transmittance_wl_grid(wl_struct *wl)
{
  int iv;
  float lambda, wvn_step_t;
  
  /* Calculate number of wavelengths needed for transmittance calculation */

  lambda = wl->wl_start;
  iv = 1;
  while (lambda < wl->wl_end) {
    if ( lambda < 350.0 ) {
      wvn_step_t  = 0.5;
    }
    else {
      wvn_step_t  = 1.0;
	}
    lambda = lambda + wvn_step_t;
    iv++;
  }
  wl->nlambda_t = iv;
  printf("Wl: %d\n",wl->nlambda_t);
  
  wl->lambda_t = (float *) calloc (wl->nlambda_t, sizeof(float));

  /* Set wavelengths for transmittance calculation */

  lambda = wl->wl_start;
  for (iv=0;iv<wl->nlambda_t;iv++) {
    wl->lambda_t[iv] = lambda;
    if ( lambda < 350.0 ) {
      wvn_step_t  = 0.5;
    }
    else {
      wvn_step_t  = 1.0;
	}
    lambda = lambda + wvn_step_t;
  }
  /* IF ( nlambda_t .GT. maxwvn ) THEN         
            WRITE(*,*)'uvspec: nlambda_t .GT. maxwvn', nlambda_t,
     $        maxwvn
            STOP
      ENDIF*/
  return 0;
}

/*************************************************************/
/* print usage information                                   */
/*************************************************************/

static void usage()
{
  fprintf (stderr, "\nConverts start and stop times for a scan into\n");
  fprintf (stderr, "solar zenith angles as a function of wavelength.\n");
  fprintf (stderr, "All options must be set.\n\n");
  fprintf (stderr, "No error checking yet, so be careful.\n\n");

  fprintf (stderr, "Usage: time2sza [-hdlbstuvw]\n");
  fprintf (stderr, " -d day of year\n");
  fprintf (stderr, " -b latiitude\n");
  fprintf (stderr, " -l longitude\n");
  fprintf (stderr, " -s start_time\n");
  fprintf (stderr, " -t end_time\n");
  fprintf (stderr, " -u start_wavelength\n");
  fprintf (stderr, " -v end_wavelength\n");
  fprintf (stderr, " -w step_wavelength\n");
  fprintf (stderr, " -h    Print this message.\n");  
}

int main(int argc, char **argv) 
{

  int status=0;
  input_struct input;
  output_struct output;

  /* Initialization */
  input.ozone_column = 450.0;
  input.day_of_year  = -999;
  input.atm.ndens    = 2;
  input.atm.nzout    = 1;
  input.atm.zout     = (float *) calloc (input.atm.nzout, sizeof(float));
  input.rte.maxcly   = 131;
  input.rte.maxwvn   = 1000;

  /*  while ((c=getopt (argc, argv, "hd:l:b:s:t:u:v:w:")) != EOF)  {
    switch(c)  {
    case 'b': 
      latitude = atof(optarg);
      break;
    case 'd': 
      day = atof(optarg);
      break;
    case 'l': 
      longitude = atof(optarg);
      break;
    case 's': 
      start_time = atof(optarg);
      break;
    case 't': 
      end_time = atof(optarg);
      break;
    case 'u': 
      start_wvn = atof(optarg);
      break;
    case 'v': 
      end_wvn = atof(optarg);
      break;
    case 'w': 
      step_wvn = atof(optarg);
      break;
    case 'h': 
      usage();
      return (-1);
      break;
    default:
      usage();
      return (-1);
    }
  }
  */
  input.wl.wl_start         = 280.0;
  input.wl.wl_end           = 420.0;
  input.day_of_year         = 172;
  input.atmosphere_filename = "../data/atmmod/afglus.dat";

  input.filepath = (char *) calloc (strlen("../data/"), sizeof(char));
  input.filepath = "../data/";

  /*  printf("%d %s %s\n", sizeof(input.filepath), input.filepath+1, input.filepath+2);
  printf("%s %s %s\n", input.filepath, input.filepath+1, input.filepath+2);

  return 0;*/

  status = uvspec(input,&output);

  printf("Ecc: %f\n",output.eccentricity_factor);

  return status;

}

int uvspec(input_struct input, output_struct *output)
{

  int iv, lev;
  int status = 0;

  float F77_FCN (eccentricity) ();
  float F77_FCN (o3column) (float* dens, int* nlyr, float* zd);
  void  F77_FCN (scaleozone) (float* dens, int* nlyr, 
			       float* ozone_column, float* zd, float* ozone_scale_factor);
  void  F77_FCN (trans) (float* aer_dtau, float* aer_gg, float* aer_ssa, 
			  float* albedo_t, float* dens,  float* dfdt, float* flup, 
			  float* ice_dtau, float* ice_gg, float* ice_ssa, 
			  float* lambda_t, int* ndens, int* nlambda_t, int* nlyr, int* nzout, 
			  const char* filepath, float* press, float* rfldir, float* rfldn, 
			  float* sza_t, float* temper, float* uavg,  float* uavgdn, 
			  float* uavgso, float* uavgup, float* wc_dtau, float* wc_gg, 
			  float* wc_ssa, float* zd, float* zout);

  char *string = "../data/";

  /* Set wavelength grid for the transmittance calculation */

  status = set_transmittance_wl_grid(&input.wl);

  /*  for(iv=0;iv<input.wl.nlambda_t;iv++){
    printf("wl: %d %f\n",iv, input.wl.lambda_t[iv]);
  }*/

  /* Correction factor for Earth-Sun distance */

  output->eccentricity_factor = 1.0;

  if ( input.day_of_year > 0.0 ) {
    output->eccentricity_factor = F77_FCN (eccentricity) ( &input.day_of_year );
  }

  /* Read in model atmosphere */

  status = read_atmosphere(input.atmosphere_filename, &input.atm);

  input.rte.nlev = input.atm.nlev;
  input.rte.nlyr = input.atm.nlyr;

  /*  for (lev=0;lev<input.rte.nlev;lev++) {
    printf("%d %f %e %e %e\n",lev,input.atm.zd[lev], input.atm.press[lev],
	   input.atm.dens[AIR][lev], input.atm.dens[O3][lev]);
  }*/
  
     /*
      IF ( nlyr .GT. maxcly ) THEN         
            WRITE(*,*)'uvspec: nlyr .GT. maxcly', nlyr, maxcly
            STOP
      ENDIF      */

  /* Scale ozone column */

  output->ozone_column = F77_FCN (o3column) ( input.atm.dens[O3], &input.rte.nlyr, input.atm.zd );
  printf("O3 before: %f\n",output->ozone_column);

  if ( input.ozone_column >= 0.0 ) {
    F77_FCN (scaleozone) ( input.atm.dens[O3], &input.atm.nlyr, &input.ozone_column,
			   input.atm.zd, &output->ozone_scale_factor );
  }
  output->ozone_column = F77_FCN (o3column) ( input.atm.dens[O3], &input.atm.nlyr, input.atm.zd );

  printf("O3 after: %f %f\n",output->ozone_column, output->ozone_scale_factor);

  /*for (lev=0;lev<input.atm.nlev;lev++) {
     printf("%d %f %e %e %e\n",lev,input.atm.zd[lev], input.atm.press[lev],
	    input.atm.dens[AIR][lev], input.atm.dens[O3][lev]);
  }*/
  
  /* Calculate radiation field with fbeam = 1.0, not exactly the same as 
     the transmittance, but no extraterrestrial spectrum involved. */
 
  if ( (status = ASCII_calloc_float (&output->flup,   input.rte.maxwvn, input.rte.maxcly)) != 0 )  return status;
  if ( (status = ASCII_calloc_float (&output->dfdt,   input.rte.maxwvn, input.rte.maxcly)) != 0 )  return status;
  if ( (status = ASCII_calloc_float (&output->rfldir, input.rte.maxwvn, input.rte.maxcly)) != 0 )  return status;
  if ( (status = ASCII_calloc_float (&output->rfldn,  input.rte.maxwvn, input.rte.maxcly)) != 0 )  return status;
  if ( (status = ASCII_calloc_float (&output->uavgdn, input.rte.maxwvn, input.rte.maxcly)) != 0 )  return status;
  if ( (status = ASCII_calloc_float (&output->uavgso, input.rte.maxwvn, input.rte.maxcly)) != 0 )  return status;
  if ( (status = ASCII_calloc_float (&output->uavgup, input.rte.maxwvn, input.rte.maxcly)) != 0 )  return status;
  output->uavg = (float *) calloc (input.rte.maxcly, sizeof(float));

  if ( (status = ASCII_calloc_float (&input.aer.dtau, input.rte.maxwvn, input.rte.maxcly)) != 0 ) return status;
  if ( (status = ASCII_calloc_float (&input.aer.gg, input.rte.maxwvn, input.rte.maxcly)) != 0 )   return status;
  if ( (status = ASCII_calloc_float (&input.aer.ssa, input.rte.maxwvn, input.rte.maxcly)) != 0 )  return status;
  if ( (status = ASCII_calloc_float (&input.ice.dtau, input.rte.maxwvn, input.rte.maxcly)) != 0 ) return status;
  if ( (status = ASCII_calloc_float (&input.ice.gg, input.rte.maxwvn, input.rte.maxcly)) != 0 )   return status;
  if ( (status = ASCII_calloc_float (&input.ice.ssa, input.rte.maxwvn, input.rte.maxcly)) != 0 )  return status;
  if ( (status = ASCII_calloc_float (&input.wc.dtau, input.rte.maxwvn, input.rte.maxcly)) != 0 ) return status;
  if ( (status = ASCII_calloc_float (&input.wc.gg, input.rte.maxwvn, input.rte.maxcly)) != 0 )   return status;
  if ( (status = ASCII_calloc_float (&input.wc.ssa, input.rte.maxwvn, input.rte.maxcly)) != 0 )  return status;
  input.atm.albedo_t = (float *) calloc (input.rte.maxwvn, sizeof(float));
  input.atm.sza_t    = (float *) calloc (input.rte.maxwvn, sizeof(float));

  printf("filepath %s\n",input.filepath);

  F77_FCN  (trans) (*input.aer.dtau, *input.aer.gg, *input.aer.ssa, 
		    input.atm.albedo_t, 
		    *input.atm.dens, *output->dfdt, *output->flup,		    
		    *input.ice.dtau, *input.ice.gg, *input.ice.ssa, 
		    input.wl.lambda_t, &input.atm.ndens, &input.wl.nlambda_t,
		    &input.atm.nlyr, &input.atm.nzout, string, /* input.filepath,*/ 
		    input.atm.press, *output->rfldir, *output->rfldn, 
		    input.atm.sza_t,  input.atm.temper, 
		    output->uavg,  *output->uavgdn, *output->uavgso, *output->uavgup,
		    *input.wc.dtau, *input.wc.gg, *input.wc.ssa, 
		    input.atm.zd, input.atm.zout);

  return 0;

}

