/************************************************************************/
/*                                                                      */
/* Copyright (C) 1997 Arve Kylling                                      */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*----------------------------------------------------------------------*/
/************************************************************************/

#ifndef __uvspec_h
#define __uvspec_h

#include <stdio.h>

/* Indices for the reading of the atmosphere files */
#define ATMFILE_ZD       0
#define ATMFILE_PRESS    1
#define ATMFILE_TEMPER   2
#define ATMFILE_AIR      3
#define ATMFILE_O3       4

#define AIR      0
#define O3       1

/* Raditive Transfer Equation STRUCTURE */
/* See disort.doc for description of variables */

typedef struct {
  int     maxcly;
  int     maxwvn;
  int     nlev;
  int     nlyr;
} rte_struct;

/* Aerosol STRUCTURE */

typedef struct {
  float   **dtau;
  float   **gg;
  float   **ssa;
} aer_struct;

/* Water Cloud STRUCTURE */

typedef struct {
  float   **dtau;
  float   **gg;
  float   **ssa;
} wc_struct;

/* Ice Cloud STRUCTURE */

typedef struct {
  float   **dtau;
  float   **gg;
  float   **ssa;
} ice_struct;

/* Wavelength grid STRUCTURE */

typedef struct {
  float   wl_start;
  float   wl_end;
  int      nlambda_t;
  float*  lambda_t;
  int      nlambda_h;
  float*  lambda_h;
} wl_struct;

/* Atmosphere data STRUCTURE */

typedef struct {
  float **dens;
  float *press;
  float *temper;
  float *zd;
  float *zout;
  float *albedo_t;
  float *sza_t;
  int    nlev;
  int    nlyr;
  int    ndens;
  int    nzout;
} atm_struct;

/* INPUT STRUCTURE */

typedef struct {
  char*        filepath;
  char*        atmosphere_filename;
  float       day_of_year;
  float       ozone_column;
  rte_struct   rte;
  atm_struct   atm;
  aer_struct   aer;
  ice_struct   ice;
  wc_struct   wc;
  wl_struct    wl;
} input_struct;


/* OUTPUT STRUCTURE */

typedef struct {
  float eccentricity_factor;
  float       ozone_column;
  float ozone_scale_factor;
  float       **flup;
  float       **dfdt;
  float       **rfldir;
  float       **rfldn;
  float       **uavgdn;
  float       **uavgso;
  float       **uavgup;
  float       *uavg;
} output_struct;

int uvspec(input_struct input, output_struct *output);











/***********************************************************/
/* system specific data: auxiliary programs, data files,   */
/*                       path names, ...                   */
/***********************************************************/

#define PROG_UVSPEC        "uvspec"
#define PROG_TUVSPEC       "tuvspec"
#define PROG_SUVSPEC       "suvspec"
#define PROG_PSUVSPEC      "psuvspec"
#define PROG_UVSPEC_ENH    "uvs_enh"
#define PROG_TUVSPEC_ENH   "tuv_enh"
#define PROG_SUVSPEC_ENH   "suv_enh"
#define PROG_PSUVSPEC_ENH  "psu_enh"


/* what to use when alpha or beta not available */
#define ALPHA_NAN          -999
#define BETA_NAN           -999

/* what to use when omega is not available */
#define OMEGA_NAN          -999

/***********************************************************/



/* MODEL is used to select the type of model: uvspec, tuvspec, ... */
typedef enum {TUVSPEC, UVSPEC, PSUVSPEC, SUVSPEC, PSPSUVSPEC,
              TUVSPEC_ENH, UVSPEC_ENH, PSUVSPEC_ENH, SUVSPEC_ENH, 
	      PSPSUVSPEC_ENH, SASHA_TWOSTR, STAR} MODEL;


/* RESOLUTION is used to specify the slit function  */
/* HIRES:  original ATLAS2 resolution               */
/* LORES2: IFU #2 slit function                     */
/* LORES3: IFU #2 slit function                     */
typedef enum {HIRES, LORES2, LORES3, SLIT} RESOLUTION;



/* SEALEVEL holds the altitude above sealevel which may be a pre-defined */
/* value or a user-defined value                                         */

typedef enum {UNDEFINED, SEA, IFU, ZUGSPITZE, HP, USER} LEVEL;

typedef struct {
  LEVEL level;
  double altitude;
} SEALEVEL;


/* struct atmospheric_data holds all data neccessary for the */
/* Arve Kylling model.                                       */

typedef struct {
  double altitude;
  SEALEVEL sealevel;
  double albedo;
  double water_content;
  double droplet_radius;
  double cloud_bottom;
  double cloud_top;
  double ozone;
  double visibility;
  double pressure;
  double alpha;
  double beta;
  double omega;
  int season;
  int vulcan;
  int cirrus;
} atmospheric_data;



/* error codes */
#define NOT_ENOUGH_DATA             -1
#define NOT_ENOUGH_CLOUD_DATA       -2
#define NOT_ENOUGH_AEROSOL_DATA     -3
#define NOT_ENOUGH_ANGSTROM_DATA    -4

#define UNKNOWN_MODEL              -10
#define ERROR_OPENING_UVSPEC_INPUT -11
#define ERROR_EXECUTING_MODEL      -12
#define ERROR_EXECUTING_AWK        -13
#define ERROR_EXECUTING_COPY       -14

#define ERROR_OZONE                -20
#define ERROR_NO_PRESSURE          -21

#define ERROR_NEW_EXTRATERRESTRIC  -30




void fprintf_atmospheric (FILE *f, atmospheric_data *atm);
int read_atmospheric_file (char *filename, atmospheric_data *atm);

int model_calculate (MODEL model, RESOLUTION resolution,
		     atmospheric_data *atm,
		     double *lambda, int n_lambda, 
		     double zenith, int day_of_year,
		     double **global, double **direct);

int better_model_calculate (MODEL model, RESOLUTION resolution,
			    char *slitfunction,
			    atmospheric_data *atm,
			    double *lambda, int n_lambda, 
			    double zenith, int day_of_year,
			    double **global, double **direct);
int model_test (void);


#endif



